/* -*- Mode: C; c-file-style: "bsd"; -*- */
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include "variables.h"
#include "prototypes.h"
#include "message.h"
#include "eval_context.h"
#include "element_output.h"
#include "identifier.h"
#include "types/variable.h"
#include "errors.h"
#include "i18n.h"


NxpDefinition* nxp_definition_get_simple(const char *name, NxpEvalContext *context)
{
    NxpIdentifier *ident;
    NxpDefinition *def;
    ident = nxp_identifier_new_simple(name);
    def = nxp_definition_get(ident, context);
    nxp_identifier_unref(ident);
    return def;
}



NxpDefinition* nxp_definition_get(NxpIdentifier *ident, NxpEvalContext *context)
{
    NxpDefinition *def;
    const char *name;
    NxpNameSpace *ns;
    g_return_val_if_fail(ident, NULL);
    nxp_identifier_resolve(ident, context, &ns, &name);
    if (ns && name)
	def = nxp_name_space_lookup_definition_r(ns, name, TRUE);
    else
	def = NULL;
    return def;
}



NxpElement* nxp_variable_get(NxpIdentifier *ident, NxpEvalContext *context)
{
    NxpElement *val;
    NxpDefinition *def;
    NxpNameSpace *ns;
    const char *name;
    g_return_val_if_fail(ident, NULL);
    nxp_identifier_resolve(ident, context, &ns, &name);
    if (ns && name) {
	nxp_name_space_lock(ns);
	def = nxp_name_space_lookup_definition(ns, name);
	if (def) {
	    if (def->type == NXP_DEF_VARIABLE)
		val = NXP_ELEMENT_REF(NXP_DEF_VARIABLE(def)->value);
	    else
		val = NULL;
	    nxp_definition_unref(def);
	} else val = NULL;
	nxp_name_space_unlock(ns);
    } else val = NULL;
    return val;
}




void nxp_user_variable_define_simple(const char *name, NxpElement *value,
				     NxpEvalContext *context, GError **error)
{
    NxpIdentifier *ident;
    GError *erro = NULL;
    ident = nxp_identifier_new_simple(name);
    nxp_user_variable_define(ident, value, context, &erro);
    if (erro) {
	g_propagate_error(error, erro);
    }
    nxp_identifier_unref(ident);
}



void nxp_user_variable_define(NxpIdentifier *ident, NxpElement *value, 
			      NxpEvalContext *context, GError **error)
{
    NxpDefinition *def;
    NxpNameSpace *ns, *actual_ns;
    const char *name;

    g_return_if_fail(ident);
    g_return_if_fail(value);
    nxp_identifier_resolve(ident, context, &ns, &name);
    g_return_if_fail(ns && name);
    nxp_name_space_lock(ns);
    if ((def = nxp_name_space_lookup_definition_extended(ns, name, &actual_ns, FALSE)))
    {
	if (def->type == NXP_DEF_VARIABLE) {
	    NXP_ELEMENT_UNREF(NXP_DEF_VARIABLE(def)->value);
	    NXP_DEF_VARIABLE(def)->value = NXP_ELEMENT_REF(value);
	} else {
		g_set_error(error, nxp_error_domain, NXP_ERROR_ALREADY_DEFINED, _("There is already a function with that name (%s)"), name);
	}
	nxp_definition_unref(def);

    } else {
	def = (NxpDefinition *)nxp_def_variable_new(g_strdup(name), value);
	nxp_name_space_add_definition(ns, def);
    }
    nxp_name_space_unlock(ns);
}






/**
 * nxp_built_in_function_define:
 * @name: name
 * @eval_func: eval function
 * @derive_func: derive function
 * @simplify_func: simplify function
 * 
 * Legacy API call to define a builtin function
 **/
void nxp_built_in_function_define(const char *name, 
				  gpointer eval_func,
				  gpointer  derive_func, 
				  gpointer  simplify_func,
				  NxpEvalContext *context)
{
    GString *new_name = g_string_new(NULL);
    NxpIdentifier *ident;
    const char *c;
    /* strip the /arity from the old style name */
    for (c = name; c && *c; c++)
	if (*c == '/') break;
	else new_name = g_string_append_c(new_name, *c);
    ident = nxp_identifier_new_simple(new_name->str);
    g_string_free(new_name, TRUE);
    nxp_bifunc_define(ident, &nxp_prototype__ELLIPSES,
		      eval_func, derive_func, simplify_func, context, NULL);
}




void nxp_variable_undefine_at_top(const char *name, NxpEvalContext *context)
{
    NxpNameSpace *ns;
    NxpDefinition *def;
    ns = context->cwd;
    g_return_if_fail(ns);
    nxp_name_space_lock(ns);
    def = nxp_name_space_lookup_definition(ns, name);
    if (def) {
	nxp_name_space_remove_definition(ns, def);
	nxp_definition_unref(def);
    }
    nxp_name_space_unlock(ns);
}




void nxp_user_function_define(NxpIdentifier *ident, NxpElement *args,
			      NxpElement *expression,
			      NxpEvalContext *context, GError **error)
{
    NxpDefUserFunc *def;
    const char *name;
    NxpNameSpace *ns;
    GSList *l, *list = NXP_ELEMENT_LIST(args)->list;
    GSList *new_names = NULL;
    GError *myerr = NULL;
    NxpElement *new_expr, *tmp_expr;

    nxp_identifier_resolve(ident, context, &ns, &name);

    tmp_expr = NXP_ELEMENT_REF(expression);
    for (l = list; l; l = l->next) {
	NxpElement *el, *new_el;
	const char *arg_name;
	char *new_name;

	el = NXP_ELEMENT(l->data);

	if (NXP_ELEMENT_GET_TYPE(el) != NXP_ELEMENT_TYPE_VARIABLE) {
	    GString *an = NULL;
	    elem2string(&an, el);
	    g_warning(_("Invalid argument name '%s'"), an->str); /* FIXME: this function should receive a GError ** */
	    g_string_free(an, TRUE);
	    return;
	}
	arg_name = nxp_identifier_to_function_argument(NXP_ELEMENT_VARIABLE(el)->ident);
	if (!arg_name) {
	    char *s = nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(el)->ident);
	    g_warning(_("NxpIdentifier '%s' is not valid as argument name"), s); /* FIXME: this function should receive a GError ** */
	    g_free(s);
	    return;
	}
	/* We create an argument that no user can enter, essential if we want to
	 * define functions with more than one variables. Example : we define
	 * f(x, y)=x^y and we ask to simplify f(y, 2).
	 * if we keep the expression as is, the simplification of f(y, 2) gives:
	 *   First step : f(y, y) = y^y
	 *   Second step : f(y, 2) = 2^2 : Very bad :-(
	 * 
	 * With f(x#, y#)=x# ^ y# : no problem :
	 *   First step : f(y, y#) = y ^ y#
	 *   Second step : f(y, 2) = y ^ 2 : Very good :-)
	 */
	new_name = g_strdup_printf("%s#", arg_name);
	new_names = g_slist_append(new_names, new_name);
	new_el = nxp_element_variable_new_simple(new_name);
	new_expr = nxp_element_substitute(tmp_expr, el, new_el, context, &myerr);
	NXP_ELEMENT_UNREF(new_el);
	if (myerr) {
	    g_warning("Problem in a substitution...\n");
	    g_propagate_error(error, myerr);
	    g_slist_free(new_names);
	    return;
	}
	NXP_ELEMENT_UNREF(tmp_expr);
	tmp_expr = new_expr;
    }
    def = nxp_def_user_func_new(g_strdup(name), tmp_expr);
    for (l = new_names; l; l = l->next)
	nxp_def_user_func_append_argument(def, l->data);
    NXP_ELEMENT_UNREF(tmp_expr);
    g_slist_free(new_names);
    nxp_name_space_add_definition_r(ns, (NxpDefinition *)def);
}
