/* -*- c -*- */
#include <gsl/gsl_linalg.h>
#include "errors.h"
#include "nxp_kernel.h"
#include "gsl.h"
#include "types/vector.h"
#include "gslpermut-prototypes.generated.h"



static NxpElement *
nxp_permutation2matrix(gsl_permutation * P, int sig)
{
    int             size;
    int             i,
                    j;
    NxpElement     *mat;

    sig = (sig >= 0) ? 1 : -1;

    size = gsl_permutation_size(P);
    mat = nxp_element_matrix_new(size, size, NULL);
    for (i = 0; i < size; i++) {
	int             pos = gsl_permutation_get(P, i);
	for (j = 0; j < size; j++) {
	    if (pos == j) {
		NXP_ELEMENT_MATRIX_ELEM_YX(NXP_ELEMENT_MATRIX(mat),
					   i, j) =
		    nxp_element_int_new(sig);
	    } else {
		NXP_ELEMENT_MATRIX_ELEM_YX(NXP_ELEMENT_MATRIX(mat),
					   i, j) = nxp_element_int_new(0);
	    }
	}
    }
    return mat;
}


/*
 */
static NxpElement *
nxp_gsl_linalg_L(NxpElementMatrix * m, ETC)
{
    NxpElement     *new;
    int             i,
                    j;
    if (m->width != m->height)
	return NULL;
    new = nxp_element_matrix_new(m->width, m->height, NULL);
    if (!new)
	return NULL;
    for (i = 0; i < m->width; i++) {
	for (j = 0; j < m->height; j++) {
	    if (i == j) {
		NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(new), i, j) =
		    nxp_element_real_new(1);
	    } else {
		if (i > j) {
		    NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(new), i,
					       j) =
			nxp_element_real_new(0);
		} else {
		    NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(new), i,
					       j) =
			NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_XY
					(m, i, j));
		}
	    }
	}
    }

    return new;
}

/*
 */
static NxpElement *
nxp_gsl_linalg_U(NxpElementMatrix * m, ETC)
{
    NxpElement     *new;
    int             i,
                    j;
    if (m->width != m->height)
	return NULL;
    new = nxp_element_matrix_new(m->width, m->height, NULL);
    if (!new)
	return NULL;
    for (i = 0; i < m->width; i++) {
	for (j = 0; j < m->height; j++) {
	    if (i < j) {
		NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(new), i, j) =
		    nxp_element_real_new(0);
	    } else {
		NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(new), i, j) =
		    NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_XY(m, i, j));
	    }
	}
    }
    return new;
}

/*
 */
static NxpElement *
nxp_gsl_linalg_LU_decomp(NxpElementMatrix * m, ETC)
{
    GError *err = NULL;
    int matrix_type, signal;
    NxpElement *el;

    gsl_permutation *P;

    if (m->width != m->height)
	return NULL;

    P = gsl_permutation_alloc(m->width);

    matrix_type = nxp_matrix_is_constant(m);
    switch (matrix_type) {
    case 0:
	gsl_permutation_free(P);
	return NULL;
    case 1:			/* only integers and reals */
	{
	    gsl_matrix     *gm;

	    el = nxp_element_vector_new(2, NULL);
	    gm = nxp_matrix_to_gsl_matrix(m, &err);
	    if (err) {
		NXP_ELEMENT_UNREF(el);
		gsl_permutation_free(P);
		g_propagate_error(error, err);
		return NULL;
	    }
	    gsl_linalg_LU_decomp(gm, P, &signal);
	    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(el), 0) =
		gsl_matrix_to_nxp_matrix(gm);
	    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(el), 1) =
		nxp_permutation2matrix(P, signal);
	    break;
	}
    case 2:			/* integers, reals and complexes */
	{
	    gsl_matrix_complex *gcm;

	    el = nxp_element_vector_new(2, NULL);
	    gcm = nxp_matrix_to_gsl_cmatrix(m, &err);
	    if (err) {
		NXP_ELEMENT_UNREF(el);
		gsl_permutation_free(P);
		g_propagate_error(error, err);
		return NULL;
	    }
	    gsl_linalg_complex_LU_decomp(gcm, P, &signal);
	    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(el), 0) =
		gsl_cmatrix_to_nxp_matrix(gcm);
	    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(el), 1) =
		nxp_permutation2matrix(P, signal);
	    break;
	}
    }
    gsl_permutation_free(P);
    GSL_RETURN(el);
}

static NxpElement *
nxp_gsl_linalg_LU_solve(NxpElementMatrix * LU,
			NxpElementPermutation * P,
			NxpElementVector * B, ETC)
{
    GError *err = NULL;
    int matrix_type;
    NxpElement *el;

    if (LU->width != LU->height)
	return NULL;

    matrix_type = nxp_matrix_is_constant(LU);
    switch (matrix_type) {
    case 0:
	return NULL;
    case 1:			/* only integers and reals */
	{
	    gsl_vector      answer;
	    gsl_matrix     *gslLU;
	    gsl_vector     *gslB;

	    gslLU = nxp_matrix_to_gsl_matrix(LU, &err);
	    if (err) { g_propagate_error(error, err); return NULL; }

	    gslB = nxp_vector_to_gsl_vector(B, context, &err);
	    if (err) { gsl_matrix_free(gslLU); g_propagate_error(error, err); return NULL; }

	    gsl_linalg_LU_solve(gslLU, P->data, gslB, &answer);

	    el = gsl_vector_to_nxp_vector(&answer);

	    gsl_vector_free(gslB);
	    gsl_matrix_free(gslLU);
	}
	break;
    case 2:			/* integers, reals and complexes */
	{
	    gsl_vector_complex answer;
	    gsl_matrix_complex *gslLU;
	    gsl_vector_complex *gslB;

	    gslLU = nxp_matrix_to_gsl_cmatrix(LU, &err);
	    if (err) { g_propagate_error(error, err); return NULL; }

	    gslB = nxp_vector_to_gsl_cvector(B, &err);
	    if (err) { 
		gsl_matrix_complex_free(gslLU); 
		g_propagate_error(error, err); return NULL; 
	    }

	    gsl_linalg_complex_LU_solve(gslLU, P->data, gslB, &answer);

	    el = gsl_cvector_to_nxp_vector(&answer);

	    gsl_vector_complex_free(gslB);
	    gsl_matrix_complex_free(gslLU);
	}
	break;
    }

    GSL_RETURN(el);
}


/*
 * - Function: int gsl_linalg_LU_refine (const gsl_matrix * A, const
 * gsl_matrix * LU, const gsl_permutation * P, const gsl_vector * B,
 * gsl_vector * X, gsl_vector * RESIDUAL) - Function: int
 * gsl_linalg_complex_LU_refine (const gsl_matrix_complex * A, const
 * gsl_matrix_complex * LU, const gsl_permutation * P, const
 * gsl_vector_complex * B, gsl_vector_complex * X, gsl_vector_complex *
 * RESIDUAL) These functions apply an iterative improvement to X, the
 * solution of A x = b, using the LU decomposition of A into (LU,P). The
 * initial residual r = A x - b is also computed and stored in RESIDUAL.
 * 
 * - Function: int gsl_linalg_LU_invert (const gsl_matrix * LU, const
 * gsl_permutation * P, gsl_matrix * INVERSE) - Function: int
 * gsl_complex_linalg_LU_invert (const gsl_matrix_complex * LU, const
 * gsl_permutation * P, gsl_matrix_complex * INVERSE) These functions
 * compute the inverse of a matrix A from its LU decomposition (LU,P),
 * storing the result in the matrix INVERSE. The inverse is computed by
 * solving the system A x = b for each column of the identity matrix.
 * 
 * - Function: double gsl_linalg_LU_det (gsl_matrix * LU, int SIGNUM) -
 * Function: gsl_complex gsl_linalg_complex_LU_det (gsl_matrix_complex *
 * LU, int SIGNUM) These functions compute the determinant of a matrix A
 * from its LU decomposition, LU. The determinant is computed as the
 * product of the diagonal elements of U and the sign of the row
 * permutation SIGNUM.
 * 
 * - Function: double gsl_linalg_LU_sgndet (gsl_matrix * LU) - Function:
 * double gsl_linalg_complex_LU_sgndet (gsl_matrix_complex * LU) ... 
 */

static NxpElement *
nxp_gsl_linalg_LU_lndet(NxpElementMatrix * LU, ETC)
{
    GError *err = NULL;
    int matrix_type;
    NxpElement *el;

    if (LU->width != LU->height)
	return NULL;

    matrix_type = nxp_matrix_is_constant(LU);
    switch (matrix_type) {
    case 0:
	return NULL;
    case 1:			/* only integers and reals */
	{
	    double res;
	    gsl_matrix *gslLU;

	    gslLU = nxp_matrix_to_gsl_matrix(LU, &err);
	    if (err) { g_propagate_error(error, err); return NULL; }

	    res = gsl_linalg_LU_lndet(gslLU);
	    el = nxp_element_real_new(res);
	    gsl_matrix_free(gslLU);
	}
	break;
    case 2:			/* integers, reals and complexes */
	{
	    double          res;
	    gsl_matrix_complex *gslLU;

	    gslLU = nxp_matrix_to_gsl_cmatrix(LU, &err);
	    if (err) { g_propagate_error(error, err); return NULL; }

	    res = gsl_linalg_complex_LU_lndet(gslLU); 
	    el = nxp_element_real_new(res);
	    gsl_matrix_complex_free(gslLU);
	}
	break;
    }
    GSL_RETURN(el);
}


void
nxp_gsl_linalg_init(NxpEvalContext * context)
{
    NxpIdentifier  *id;
nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"linearAlgebra");
id = nxp_identifier_new_with_full_path(3,"gsl","linearAlgebra", "L");
 nxp_bifunc_define(id, &nxp_prototype__MATRIX, nxp_gsl_linalg_L, NULL, nxp_gsl_linalg_L, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","linearAlgebra", "U");
 nxp_bifunc_define(id, &nxp_prototype__MATRIX, nxp_gsl_linalg_U, NULL, nxp_gsl_linalg_U, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","linearAlgebra", "LU");
 nxp_bifunc_define(id, &nxp_prototype__MATRIX, nxp_gsl_linalg_LU_decomp, NULL, nxp_gsl_linalg_LU_decomp, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","linearAlgebra", "LUsolve");
 nxp_bifunc_define(id, &gslpermut_prototype__MATRIX_PERMUTATION_VECTOR, nxp_gsl_linalg_LU_solve, NULL, nxp_gsl_linalg_LU_solve, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","linearAlgebra", "LULnDet");
 nxp_bifunc_define(id, &nxp_prototype__MATRIX, nxp_gsl_linalg_LU_lndet, NULL, nxp_gsl_linalg_LU_lndet, context, NULL); nxp_identifier_unref(id);
}
