/* -*- c -*- */
#include <gsl/gsl_roots.h>
#include "nxp_kernel.h"
#include "named_args.h"
#include "gsl.h"
#include "i18n.h"


typedef struct _eval_data {
    NxpEvalContext *context;
    NxpIdentifier *identifier;
    NxpIdentifier *identifier_d;
} EvalData;

static double eval_nxp_def_function(double x, void *data, gboolean d) {
    GError *err = NULL;
    NxpElement *function;
    NxpIdentifier *id;
    NxpElement *answer;
    EvalData* ed = (EvalData*)data;
    double val = NAN;
    GSList *l;
    NxpElement *args;


    id = d ? ed->identifier_d : ed->identifier;
    l = g_slist_append(NULL, nxp_element_real_new(x));
    args = nxp_element_list_new_from_list(l);

    function = nxp_element_function_new(id, args, NULL);
    NXP_ELEMENT_UNREF(args);

    if (!function) return NAN;

    answer = nxp_element_eval(function, ed->context, &err);
    NXP_ELEMENT_UNREF(function);
    if (!answer || err) return NAN;

    if (NXP_ELEMENT_GET_TYPE(answer) == NXP_ELEMENT_TYPE_REAL)
	val = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(answer));
    if (NXP_ELEMENT_GET_TYPE(answer) == NXP_ELEMENT_TYPE_INT)
	val = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(answer));

    NXP_ELEMENT_UNREF(answer);
    return val;
}

static double eval_f(double x, void *data) {
    return eval_nxp_def_function(x, data, FALSE);
}

static double eval_df(double x, void *data) {
    return eval_nxp_def_function(x, data, TRUE);
}


static void eval_fdf(double x, void *data, double *f, double *df) {
    *f  = eval_nxp_def_function(x, data, FALSE);
    *df = eval_nxp_def_function(x, data, TRUE);
}

static NxpElement* nxp_gsl_roots_brack_generic(const gsl_root_fsolver_type *gsl_solver,
					       NxpElementVariable *var,
					       NxpElementReal *low,
					       NxpElementReal *high, ETC)
{
    NxpIdentifier *func_id, *ufunc_id = NULL;
    NxpDefinition *func_def;
    EvalData data;
    gsl_function function;
    gsl_root_fsolver *solver;

    func_id = var->ident;
    func_def = nxp_definition_get(func_id, context);

    if (!func_def) {
	ufunc_id = nxp_identifier_create_function_identifier(func_id, 1);
	func_def = nxp_definition_get(ufunc_id, context);
	nxp_identifier_unref(ufunc_id);
    }
    
    if (func_def && (func_def->type == NXP_DEF_USER_FUNC ||
		     func_def->type == NXP_DEF_BI_FUNC)) {
	double val;
	double last_iter = 0;
	GError *err = NULL;
	NxpElementReal *nxp_delta = NULL;
	NxpElementInt *nxp_iters = NULL;
	int iters = 0;
	gboolean iters_def = FALSE;
	gboolean delta_def = FALSE;
	gboolean gsl_error = FALSE;
	double cdelta = 0;
	double delta = 0;

	/* Initialize the solver */
	solver = gsl_root_fsolver_alloc(gsl_solver);
	if (!solver) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_MEMORY,
			_("Not enough memory to start solver engine"));
	    return NULL;
	}

	data.identifier = nxp_identifier_ref(func_id);
	data.identifier_d = NULL;
	data.context = context;

	/* GET DELTA */
	nxp_delta = nxp_named_args_get_real(named_args, "delta",
					    context, &err);
	if (err) err = NULL;
	if (nxp_delta) delta = nxp_element_r_get_value_as_double(nxp_delta);

	/* GET TERACTIONS */
	nxp_iters = nxp_named_args_get_int(named_args, "iters", 
					   context, &err);
	if (nxp_iters) iters = nxp_element_r_get_value_as_int(nxp_iters);
	
	function.function = &eval_f;
	function.params   = &data;
	
	/* Initialize the solver */
	if (gsl_root_fsolver_set(solver, &function,
				 nxp_element_r_get_value_as_double(low),
				 nxp_element_r_get_value_as_double(high))) 
	    gsl_error = TRUE;
	
	/* We can't have both values undefined... */
	if (iters==0 && delta==0.0000000) iters = 10;

	/* Set boolean flags if we have any of them to check...  This
	   could be done another way, but, it is working! */
	if (iters) iters_def = TRUE;
	if (delta) delta_def = TRUE;

	/* Current delta is zero. If delta is zero and delta_def true,
	   then, it is the first iteraction. */
	cdelta = 0;
	
	do {
	    if (gsl_root_fsolver_iterate(solver)) gsl_error = TRUE;
	    val = gsl_root_fsolver_root(solver);

	    /* Decrement iteractions, if we are using them */
	    if (iters_def) iters--;
	    
	    /* Calculate current delta value */
	    if (delta_def) {
		if (cdelta == 0.0000000)  cdelta = delta+1;
		else		          cdelta = ABS(last_iter - val);
		last_iter = val;
	    }
	} while (gsl_error!=TRUE && 
		 ((iters_def && iters) || !iters_def) &&
		 ((delta_def && cdelta > delta) || !delta_def));
	
	gsl_root_fsolver_free(solver);
	
	nxp_identifier_unref(func_id);

	GSL_RETURN(nxp_element_real_new(val));
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED, 
		    _("Undefined function"));
	return NULL;
    }
}


static NxpElement* nxp_gsl_roots_falsepos(NxpElementVariable *var,
					  NxpElementReal *low,
					  NxpElementReal *high, ETC) {
    return nxp_gsl_roots_brack_generic(gsl_root_fsolver_falsepos,
				       var, low, high, named_args, context, error);
}

static NxpElement* nxp_gsl_roots_bisection(NxpElementVariable *var,
					  NxpElementReal *low,
					  NxpElementReal *high, ETC) {
    return nxp_gsl_roots_brack_generic(gsl_root_fsolver_bisection,
				       var, low, high, named_args, context, error);
}

static NxpElement* nxp_gsl_roots_brent(NxpElementVariable *var,
				       NxpElementReal *low,
				       NxpElementReal *high, ETC) {
    return nxp_gsl_roots_brack_generic(gsl_root_fsolver_brent,
				       var, low, high, named_args, context, error);
}



static NxpElement* nxp_gsl_roots_derivative_generic(const gsl_root_fdfsolver_type *gsl_solver,
						    NxpElementVariable *f,
						    NxpElementVariable *df,
						    NxpElementReal *aprox,
						    ETC)
{
    NxpIdentifier *f_id, *uf_id;
    NxpIdentifier *df_id, *udf_id;
    NxpDefinition *f_def;
    NxpDefinition *df_def;
    EvalData data;
    gsl_root_fdfsolver *solver; 

    gsl_function_fdf function;

    f_id = f->ident;
    df_id = df->ident;

    f_def = nxp_definition_get(f_id, context);
    if (!f_def) {
 	uf_id = nxp_identifier_create_function_identifier(f_id, 1);
 	f_def = nxp_definition_get(uf_id, context);
 	nxp_identifier_unref(uf_id);
    }

    df_def = nxp_definition_get(df_id, context);
    if (!df_def) {
 	udf_id = nxp_identifier_create_function_identifier(df_id, 1);
 	df_def = nxp_definition_get(udf_id, context);
 	nxp_identifier_unref(udf_id);
    }
    
    if (f_def && (f_def->type == NXP_DEF_USER_FUNC || f_def->type == NXP_DEF_BI_FUNC)) {
	if (df_def && (df_def->type == NXP_DEF_USER_FUNC || df_def->type == NXP_DEF_BI_FUNC)) {
	    double val;
	    double last_iter = 0;
	    GError *err = NULL;
	    NxpElementReal *nxp_delta = NULL;
	    NxpElementInt *nxp_iters = NULL;
	    int iters = 0;
	    gboolean iters_def = FALSE;
	    gboolean delta_def = FALSE;
	    gboolean gsl_error = FALSE;
	    double cdelta = 0; 
	    double delta = 0;
	    
	    /* Initialize the solver */
	    solver = gsl_root_fdfsolver_alloc(gsl_solver);
	    if (!solver) {
		g_set_error(error, nxp_error_domain, NXP_ERROR_NO_MEMORY,
			    _("Not enough memory to start solver engine")); 
		return NULL; 
	    }
	    
	    data.identifier = f_id;
	    data.identifier_d = df_id;
	    data.context = context;

	    /* GET DELTA */
	    nxp_delta = nxp_named_args_get_real(named_args, "delta", context, &err);
	    if (err) err = NULL;
	    if (nxp_delta) delta = nxp_element_r_get_value_as_double(nxp_delta);

	    /* GET ITERACTIONS */
	    nxp_iters = nxp_named_args_get_int(named_args, "iters", context, &err);
	    if (nxp_iters) iters = nxp_element_r_get_value_as_int(nxp_iters);
	
	    function.f = &eval_f;
	    function.df = &eval_df;
	    function.fdf = &eval_fdf;
	    function.params = &data;
	
	    g_warning("Aprox is %G", nxp_element_r_get_value_as_double(aprox));
	    
	    /* Initialize the solver */
	    if (gsl_root_fdfsolver_set(solver, &function, nxp_element_r_get_value_as_double(aprox)))
		gsl_error = TRUE;
	
	    /* We can't have both values undefined... */
	    if (iters==0 && delta==0.0000000) iters = 10;

	    /* Set boolean flags if we have any of them to check...  This
	       could be done another way, but, it is working! */
	    if (iters) iters_def = TRUE;
	    if (delta) delta_def = TRUE;

	    /* Current delta is zero. If delta is zero and delta_def true, 
	       then, it is the first iteraction. */
	    cdelta = 0;
	
	    do {
		if (gsl_root_fdfsolver_iterate(solver)) gsl_error = TRUE;
		val = gsl_root_fdfsolver_root(solver);

		g_message("Value is %G", val);
		
		/* Decrement iteractions, if we are using them */
		if (iters_def) iters--;
		
		/* Calculate current delta value */
		if (delta_def) {
		    if (cdelta == 0.0000000)  cdelta = delta+1;
		    else		      cdelta = ABS(last_iter - val);
		    last_iter = val;
		}
	    } while (gsl_error!=TRUE &&
		     ((iters_def && iters) || !iters_def) &&
		     ((delta_def && cdelta > delta) || !delta_def));
	    
	    gsl_root_fdfsolver_free(solver); 
	    
	    GSL_RETURN(nxp_element_real_new(val));
	} else { 
	    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
			_("Undefined derivative function"));
	    return NULL;
	}
    } else { 
 	g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
 		    _("Undefined function"));
 	return NULL;
    }
}


static NxpElement* nxp_gsl_roots_newton(NxpElementVariable *f,
					NxpElementVariable *df,
					NxpElementReal *aprox, ETC) {
    return nxp_gsl_roots_derivative_generic(gsl_root_fdfsolver_newton,
					    f, df, aprox, named_args, context, error);
}

static NxpElement* nxp_gsl_roots_secant(NxpElementVariable *f,
					NxpElementVariable *df,
					NxpElementReal *aprox, ETC) {
    return nxp_gsl_roots_derivative_generic(gsl_root_fdfsolver_secant,
					    f, df, aprox, named_args, context, error);
}

static NxpElement* nxp_gsl_roots_steffenson(NxpElementVariable *f,
					    NxpElementVariable *df,
					    NxpElementReal *aprox, ETC) {
    return nxp_gsl_roots_derivative_generic(gsl_root_fdfsolver_steffenson,
					    f, df, aprox, named_args, context, error);
}

void
nxp_gsl_roots_init(NxpEvalContext *context)
{
    NxpIdentifier  *id;
nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"rootFinding");
id = nxp_identifier_new_with_full_path(3,"gsl","rootFinding", "falsepos");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_REAL_REAL, nxp_gsl_roots_falsepos, NULL, nxp_gsl_roots_falsepos, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","rootFinding", "bisection");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_REAL_REAL, nxp_gsl_roots_bisection, NULL, nxp_gsl_roots_bisection, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","rootFinding", "brent");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_REAL_REAL, nxp_gsl_roots_brent, NULL, nxp_gsl_roots_brent, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","rootFinding", "newton");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VARIABLE_REAL, nxp_gsl_roots_newton, NULL, nxp_gsl_roots_newton, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","rootFinding", "secant");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VARIABLE_REAL, nxp_gsl_roots_secant, NULL, nxp_gsl_roots_secant, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","rootFinding", "steffenson");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VARIABLE_REAL, nxp_gsl_roots_steffenson, NULL, nxp_gsl_roots_steffenson, context, NULL); nxp_identifier_unref(id);
}

