#include <stdio.h>

#include "nxp_kernel.h"
#include "functions.h"
#include "config.h"
#include "i18n.h"

typedef enum {
    NXP_SAMPLE_STATUS_OK,       /* sample is ok */
    NXP_SAMPLE_STATUS_COMPLEX,  /* sample is a complex, skip this
                                 * sample but try to warn the user */
    NXP_SAMPLE_STATUS_IGNORE,   /* function is not defined in this
                                 * point, but keep sampling */
    NXP_SAMPLE_STATUS_ABORT     /* serious error detected, abort  */
} NxpSampleStatus;

/**
 * nxp_sample_error_p: - check for non-trivial errors during sampling
 * @error: a #GError that may have been raised by a previous
 * nxp_element_eval() call
 * @x: X value for the sample
 * @elem: a pointer to a variable containing a #NxpElement that may
 * have been obtained by a previous nxp_element_eval() call.  The
 * element may have to be converted inplace.
 * 
 * Checks if an error is important enough to abort a sampling process.
 * Currently, DivideByZero and Domain errors are ignored, while
 * everything else is reported.  It also checks a value to see if it
 * is scalar and not complex.  The element is also converted if
 * possible to real.
 * 
 * Return value: See #NxpSampleStatus
 **/
static NxpSampleStatus inline
nxp_sample_error_p(GError **error,
                   nxp_real_t x,
                   NxpElement **elem)
{
    if (*error) {
        if ((*error)->domain == nxp_error_domain) {
            if ((*error)->code == NXP_ERROR_DOMAIN
                || (*error)->code == NXP_ERROR_DIVIDE_BY_ZERO)
            {
                g_clear_error(error);
                NXP_ELEMENT_XUNREF(*elem);
                *elem = nxp_element_void_new();
                return NXP_SAMPLE_STATUS_IGNORE;
            } else
                return NXP_SAMPLE_STATUS_ABORT;
        } else
            return NXP_SAMPLE_STATUS_ABORT;
    }

    if (!NXP_ELEMENT_IS_REAL(*elem)) {
        NxpElement *real_elem;

          /* complex values are ignored */
        if (NXP_ELEMENT_IS_COMPLEX(*elem)) {
            NXP_ELEMENT_XUNREF(*elem);
            *elem = nxp_element_void_new();
            return NXP_SAMPLE_STATUS_COMPLEX;
        }
        real_elem = nxp_typecasts_try_cast(*elem, NXP_ELEMENT_TYPE_REAL);
        if (real_elem == NULL) {
            g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK,
                        _("expression yielded a result of non-scalar type (%s) at %LG"),
                        NXP_ELEMENT_GET_TYPE_NAME(elem), x);
            return NXP_SAMPLE_STATUS_ABORT;
        }
          /* now swap the elements */
        NXP_ELEMENT_UNREF(*elem);
        *elem = real_elem;
    }

    return NXP_SAMPLE_STATUS_OK;
}

/* sample[expression, variable, range_start, range_end, samples]
   Returns a vector of SAMPLES samples from EXPRESSION where VARIABLE
   takes values in the range RANGE_START to RANGE_END */
NxpElement* stdlib_eval_sample(NxpElement *el,
			       NxpElementVariable *var,
			       NxpElementReal *r_start,
			       NxpElementReal *r_end,
			       NxpElementInt *n, ETC)
{
    NxpElementVector *result;
    NxpElement *expression, *tmp;
    NxpIdentifier *variable;
    gint samples, i;
    double range_start, range_end, t, k;
    GError *err = NULL;

    expression = el;
    variable = var->ident;
    range_start = mpfr_get_d(r_start->value, GMP_RNDN);
    range_end = mpfr_get_d(r_end->value, GMP_RNDN);

    if (range_start > range_end) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("function 'sample': value error: RANGE START > RANGE END."));
	return NULL;
    }

    samples = nxp_element_int_get_si(NXP_ELEMENT(n), &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }

    if (samples < 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("argument 5 of function 'sample' (SAMPLES) should not be negative."));
	return NULL;
    }

    result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(samples, NULL));
    if (samples == 0)
	return NXP_ELEMENT(result);
    /* create anonymous name space so that variables defined here will
       not interfere with existing ones. */
    nxp_eval_context_push_name_space(context);
    k = (range_end - range_start) / (samples - 1);
    for (i = 0; i < samples; i++)
    {
	NxpElement *sample;
	GError *error1 = NULL;
	t = range_start + k * i;
	tmp = nxp_element_real_new(t);
	nxp_user_variable_define(variable, tmp, context, &error1);
	NXP_ELEMENT_UNREF(tmp);
	if (error1) {
	    NXP_ELEMENT_UNREF(result);
	    nxp_eval_context_pop_name_space(context);
	    g_propagate_error(error, error1);
	    return NULL;
	}
	sample = nxp_element_eval(expression, context, &error1);
	if (error1) {
	    NXP_ELEMENT_XUNREF(sample);
	    NXP_ELEMENT_UNREF(result);
	    nxp_eval_context_pop_name_space(context);
	    g_propagate_error(error, error1);
	    return NULL;
	}
	_element_vector_index(result, i) =  sample;
    }
    nxp_eval_context_pop_name_space(context);
    return NXP_ELEMENT(result);
}


typedef struct {
    NxpElement *x;
    NxpElement *y;
} sample_t;

NxpElement*
stdlib_eval_adaptive_sample(NxpElement *expr,
                            NxpElementVariable *var,
                            NxpElementReal *nxp_xmin,
                            NxpElementReal *nxp_xmax,
                            NxpElementReal *nxp_ymin,
                            NxpElementReal *nxp_ymax,
                            NxpElementInt *nxp_precision,
                            ETC)
{
    GArray *list;
    GError *error1 = NULL;
    nxp_real_t xmin = mpfr_get_d(nxp_xmin->value, GMP_RNDN);
    nxp_real_t xmax = mpfr_get_d(nxp_xmax->value, GMP_RNDN);
    nxp_real_t ymin = mpfr_get_d(nxp_ymin->value, GMP_RNDN);
    nxp_real_t ymax = mpfr_get_d(nxp_ymax->value, GMP_RNDN);
    int precision = nxp_element_int_get_si(NXP_ELEMENT(nxp_precision), &error1);
    const int initial_samples = 100;
    nxp_real_t k, t, t1 = 0;
    int i;
    NxpIdentifier *variable = var->ident;
    nxp_real_t minimum = (xmax - xmin) / precision;
    nxp_real_t eps = (ymax - ymin) / precision;
    NxpElement *retval;
    enum y_pos {Y_POS_UNSET,
                Y_POS_BELOW,
                Y_POS_INSIDE,
                Y_POS_ABOVE} prev_y_pos = Y_POS_UNSET;

    if (error1) {
	g_propagate_error(error, error1);
	return NULL;
    }

#define get_y_pos(yr) \
    ((yr < ymin)? Y_POS_BELOW : ((yr > ymax)? Y_POS_ABOVE : Y_POS_INSIDE))


    if (xmin >= xmax) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("adaptive_sample: value error: XMIN >= XMAX."));
	return NULL;
    }

    if (ymin >= ymax) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("adaptive_sample: value error: YMIN >= YMAX."));
	return NULL;
    }

    if (precision < 1) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("adaptive_sample: value error: PRECISION >= 1"));
	return NULL;
    }

    list = g_array_new(FALSE, FALSE, sizeof(sample_t));

    /* create anonymous name space so that variables defined here will
       not interfere with existing ones. */
    nxp_eval_context_push_name_space(context);

      /* we start by doing linear sampling, and later adaptively add
       * more samples */
    k = (xmax - xmin) / (initial_samples - 1);
    for (i = 0; i < initial_samples; i++)
    {
	NxpElement *x, *y;
	t = xmin + k * i;
	x = nxp_element_real_new(t);
	nxp_user_variable_define(variable, x, context, &error1);
	y = nxp_element_eval(expr, context, &error1);

        switch (nxp_sample_error_p(&error1, t, &y))
        {
        case NXP_SAMPLE_STATUS_COMPLEX: /* TODO: warn user of complex value */
        case NXP_SAMPLE_STATUS_IGNORE: {
            sample_t s = { x, y };
            prev_y_pos = Y_POS_UNSET;
            g_array_append_val(list, s);
            break;
        }

        case NXP_SAMPLE_STATUS_OK: {
            nxp_real_t yr = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(y));
            enum y_pos new_y_pos;
            sample_t s = { x, y };

            new_y_pos = get_y_pos(yr);
            if (   (new_y_pos == Y_POS_BELOW && prev_y_pos == Y_POS_ABOVE)
                || (new_y_pos == Y_POS_ABOVE && prev_y_pos == Y_POS_BELOW))
            {
                  /* likely that a discontinuity occurred -> insert a void value */
                sample_t s = { nxp_element_real_new((t1 + t)/2),
                               nxp_element_void_new() };
                g_array_append_val(list, s);
            }
            prev_y_pos = new_y_pos;
            g_array_append_val(list, s);
            break;
        }

        case NXP_SAMPLE_STATUS_ABORT:
            NXP_ELEMENT_UNREF(x);
            NXP_ELEMENT_XUNREF(y);
            goto error;

        }
        t1 = t;
    }

    minimum = (xmax - xmin)/precision;
    eps = (ymax - ymin)/precision;
    i = 0;
    while (i < list->len - 1)
    {
        nxp_real_t a_x, a_y;
        nxp_real_t b_x, b_y;
        nxp_real_t x, y;
        sample_t *samp_a, *samp_b;
        /* a = L[i] */
        samp_a = &g_array_index(list, sample_t, i);
        samp_b = &g_array_index(list, sample_t, i + 1);
        a_x = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(samp_a->x));
        b_x = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(samp_b->x));

        if (!samp_a->y || NXP_ELEMENT_IS_VOID(samp_a->y))
        {
            sample_t s;

            if (!samp_b->y || NXP_ELEMENT_IS_VOID(samp_b->y)) {
                  /* if both points are undefined, jump to next pair */
                ++i;
                continue;
            }
              /* We've hit a discontinuity! Now we iterate backward
               * trying to make smaller jumps to find out more
               * accurately where this discontinuity happens.  */
            x = (a_x + b_x)/2;
            if (b_x - x <= minimum) {
                ++i; /* stop searching for optimum discontinuity point */
                continue;
            }
            s.x = nxp_element_real_new(x);
            nxp_user_variable_define(variable, s.x, context, &error1);
            g_assert(error1 == NULL);
            s.y = nxp_element_eval(expr, context, &error1);

            switch (nxp_sample_error_p(&error1, x, &s.y))
            {
            case NXP_SAMPLE_STATUS_COMPLEX: /* TODO: warn user of complex value */
            case NXP_SAMPLE_STATUS_IGNORE:
                g_array_insert_val(list, i + 1, s);
                break;

            case NXP_SAMPLE_STATUS_OK: {
                enum y_pos new_y_pos;
                nxp_real_t yr = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(s.y));
                prev_y_pos = get_y_pos(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(samp_b->y)));
                new_y_pos = get_y_pos(yr);
                if ((new_y_pos == Y_POS_BELOW && prev_y_pos == Y_POS_ABOVE) ||
                    (new_y_pos == Y_POS_ABOVE && prev_y_pos == Y_POS_BELOW))
                {
                      /* likely that a discontinuity occurred -> insert a void value */
                    sample_t s1 = { nxp_element_real_new((x + b_x)/2),
                                    nxp_element_void_new() };
                    g_array_insert_val(list, i + 1, s1);
                    NXP_ELEMENT_UNREF(s.x);
                    NXP_ELEMENT_UNREF(s.y);
                    ++i;
                } else
                    g_array_insert_val(list, i + 1, s);
                break;
            }

            case NXP_SAMPLE_STATUS_ABORT:
                NXP_ELEMENT_UNREF(s.x);
                goto error;
            }
            continue;
        }
        a_y = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(samp_a->y));

        /* b = L[i + 1] */
        if (!samp_b->y || NXP_ELEMENT_IS_VOID(samp_b->y))
        {
              /* We've hit a discontinuity! Now we iterate forward
               * trying to make smaller jumps to find out more
               * accurately where this discontinuity happens.  */
            sample_t s;

            x = (a_x + b_x)/2;
            if (x - a_x <= minimum) {
                ++i; /* stop searching for optimum discontinuity point */
                continue;
            }
            s.x = nxp_element_real_new(x);
            nxp_user_variable_define(variable, s.x, context, &error1);
            g_assert(error1 == NULL);
            s.y = nxp_element_eval(expr, context, &error1);

            switch (nxp_sample_error_p(&error1, x, &s.y))
            {
            case NXP_SAMPLE_STATUS_COMPLEX: /* TODO: warn user of complex value */
            case NXP_SAMPLE_STATUS_IGNORE:
                g_array_insert_val(list, i + 1, s);
                break;

            case NXP_SAMPLE_STATUS_OK: {
                enum y_pos new_y_pos;
                nxp_real_t yr = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(s.y));
                prev_y_pos = get_y_pos(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(samp_a->y)));
                new_y_pos = get_y_pos(yr);
                if ((new_y_pos == Y_POS_BELOW && prev_y_pos == Y_POS_ABOVE) ||
                    (new_y_pos == Y_POS_ABOVE && prev_y_pos == Y_POS_BELOW))
                {
                      /* likely that a discontinuity occurred -> insert a void value */
                    sample_t s1 = { nxp_element_real_new((a_x + x)/2),
                                    nxp_element_void_new() };
                    g_array_insert_val(list, i + 1, s1);
                    NXP_ELEMENT_UNREF(s.x);
                    NXP_ELEMENT_UNREF(s.y);
                    ++i;
                } else
                    g_array_insert_val(list, i + 1, s);
                break;
            }

            case NXP_SAMPLE_STATUS_ABORT:
                NXP_ELEMENT_UNREF(s.x);
                goto error;
            }
            continue;
        }

        b_y = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(samp_b->y));

        if ((b_x - a_x) > minimum)
        {
            NxpElement *xelem, *yelem;
            x = (a_x + b_x)/2;

              /* f1, d = f(x) */
            xelem = nxp_element_real_new(x);
            nxp_user_variable_define(variable, xelem, context, &error1);
            yelem = nxp_element_eval(expr, context, &error1);
            /* if a[1] and b[1] and d: */

            switch (nxp_sample_error_p(&error1, x, &yelem))
            {
            case NXP_SAMPLE_STATUS_COMPLEX: /* TODO: warn user of complex value */
            case NXP_SAMPLE_STATUS_IGNORE:
            {
                  /* L.insert(i + 1, [x, f1, d]) */
                sample_t s = {xelem, yelem};
                g_array_insert_val(list, i + 1, s);
                break;
            }

            case NXP_SAMPLE_STATUS_OK:
            {
                nxp_real_t ymid = (a_y + b_y) / 2;
                y = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(yelem));
                if (fabs(ymid - y) > eps) {
                    enum y_pos new_y_pos;
                    prev_y_pos = get_y_pos(a_y);
                    new_y_pos = get_y_pos(y);
                    if ((new_y_pos == Y_POS_BELOW && prev_y_pos == Y_POS_ABOVE) ||
                        (new_y_pos == Y_POS_ABOVE && prev_y_pos == Y_POS_BELOW))
                    {
                          /* likely that a discontinuity occurred -> insert a void value */
                        sample_t s1 = { nxp_element_real_new((a_x + x)/2),
                                        nxp_element_void_new() };
                        g_array_insert_val(list, i + 1, s1);
                        NXP_ELEMENT_UNREF(xelem);
                        NXP_ELEMENT_UNREF(yelem);
                        ++i;
                    } else {
                          /* L.insert(i + 1, [x, f1, d]) */
                        sample_t s = {xelem, yelem};
                        g_array_insert_val(list, i + 1, s);
                    }
                } else {
                    ++i;
                    NXP_ELEMENT_XUNREF(xelem);
                    NXP_ELEMENT_XUNREF(yelem);
                }
                break;
            }
            case NXP_SAMPLE_STATUS_ABORT:
                NXP_ELEMENT_UNREF(xelem);
                goto error;

            }

	} else {
            ++i;
        }
    }
    nxp_eval_context_pop_name_space(context);
    
    retval = nxp_element_vector_new(list->len*2, (NxpElement **) list->data);
    g_array_free(list, FALSE);
    return retval;

error:
    for (i = 0; i < list->len; ++i) {
        sample_t *samp = &g_array_index(list, sample_t, i);
        NXP_ELEMENT_UNREF(samp->x);
        NXP_ELEMENT_XUNREF(samp->y);
    }
    g_array_free(list, TRUE);
    nxp_eval_context_pop_name_space(context);
    *error = error1;
    return NULL;
#undef get_new_y_pos
}

