/* -*- Mode: C; c-file-style: "bsd" -*- */
#include "config.h"
#include <pthread.h>
#include "nxp_kernel.h"
#include "parser.h"
#include <readline.h>
#include <history.h>
#include <popt.h>
#include <locale.h>

GCompletion	*function_completion		= NULL;
char 		*command_line_eval_string 	= NULL;
gboolean	 debug_noeval			= FALSE;
gboolean	 debug_fatal_warnings		= FALSE;
char 		*rcfile 			= NULL;


/* This is defined to make it work in standalone mode */
void Numexp_Kernel_to_Client_cwd_change(NxpEvalContext *ctxt) {
}

static struct poptOption options[] = {
    { 
	"debug-parser",
	'\0',
	POPT_ARG_NONE,
	&expdebug,
	0,
	"Enable parser debugging messages",
	NULL
    },
    { 
	"debug-noeval",
	'\0',
	POPT_ARG_NONE,
	&debug_noeval,
	0,
	"Debug: do not evaluate input",
	NULL
    },
    { 
	"debug-fatal-warnings",
	'\0',
	POPT_ARG_NONE,
	&debug_fatal_warnings,
	0,
	"Debug: make all warnings fatal",
	NULL
    },
    { 
	"debug-trace-elements",
	'\0',
	POPT_ARG_NONE,
	&nxp_element_trace_references,
	0,
	"Trace element references (debug)",
	NULL
    },
    { 
	"eval",
	'e',
	POPT_ARG_STRING,
	&command_line_eval_string,
	1,
	"Evaluate expression, print result and exit",
	"EXPRESSION"
    },
    { 
	"rc",
	'r',
	POPT_ARG_STRING,
	&rcfile,
	1,
	"Startup file to load instead of $HOME/.numexprc",
	"FILE"
    },
    POPT_AUTOHELP
    {
	NULL,
	'\0',
	0,
	NULL,
	0,
	NULL,
	NULL
    }
};




/*
 * An utility function taken from readline's manual (with some changes): 
 */
/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *rl_gets(const char *prompt)
{
    /* A static variable for holding the line. */
    static char *line_read = NULL;
    /* If the buffer has already been allocated, return the memory to
       the free pool. */
    if (line_read) {
	free(line_read);
	line_read = NULL;
    }
    /* Get a line from the user. */
    line_read = readline((char *)prompt);
    /* If the line has any text in it, save it on the history. */
    if (line_read && *line_read)
	add_history(line_read);
    return line_read;
}

NxpEvalContext *main_context;

/* ================= MAIN =================== */
int main(int argc, char *argv[], char *envp[])
{
    int c;
    char *line;
    poptContext pctxt;

    /* enable i18n */
    setlocale(LC_ALL, "");
    setlocale(LC_NUMERIC, "C");

    /* ------ popt stuff --------- */
    pctxt = poptGetContext("numexpc", argc, (const char **)argv, options, 0);
    poptReadDefaultConfig(pctxt, 0);
    while ( (c = poptGetNextOpt(pctxt)) > 0 );
    if (c < -1) {
	fprintf(stderr, "%s: %s\n",
		poptBadOption(pctxt, POPT_BADOPTION_NOALIAS),
		poptStrerror(c));
	poptPrintUsage(pctxt, stderr, 0);
	return 1;
    }

    if (debug_fatal_warnings)
	g_log_set_always_fatal(G_LOG_LEVEL_ERROR|G_LOG_LEVEL_WARNING);

    /* ===== initialize NumExp ===== */
    nxp_kernel_init();
    main_context = nxp_eval_context_new(rcfile);

    nxp_element_preferences.number_precision = 12; // this is only temporary

    rl_bind_key('\t', rl_insert); // no autocompletion just yet
    
    /* ======== handle --eval ========= */
    if (command_line_eval_string)
    {
	NxpElement *element, *el1;
	GString *result;
	GError *error;

	element = exp_parse_string(command_line_eval_string, main_context, &error);
	if (error) {
	    printf("pare error: %s\n", error->message);
	    g_error_free(error);
	    return 2;
	}
	if (!element) return 2;
	error = NULL;
	el1 = nxp_element_eval(element, main_context, &error);
	NXP_ELEMENT_UNREF(element);
	if (error) {
	    printf("eval error: %s\n", error->message);
	    g_error_free(error);
	    return 2;
	}
	if (!el1) return 3;
	if (NXP_ELEMENT_GET_TYPE(el1) != NXP_ELEMENT_TYPE_VOID)
	{
	    if ( !(result = nxp_element_repr(el1))) {
		NXP_ELEMENT_UNREF(el1);
		return 3;
	    }
	    printf("ans = %s\n", result->str);
	    g_string_free(result, TRUE);
	}
	NXP_ELEMENT_UNREF(el1);
	return 0;
    }

    poptFreeContext(pctxt);

    
    /* ---------- read-eval-print cycle ---------- */
    do {
	NxpElement *element, *el1;
	GString *result;
	GError *error = NULL;

	nxp_name_space_wakeup_listeners(main_context);
	line = rl_gets("numexp> ");
	if (!line) break;

	element = exp_parse_string(line, main_context, &error);
	if (error) {
	    printf("parse error: %s\n", error->message);
	    g_clear_error(&error);
	    continue;
	}
	if (!element) continue;
	
	if (debug_noeval)
	    el1 = NXP_ELEMENT_REF(element);
	else
	    el1 = nxp_element_eval(element, main_context, &error);
	NXP_ELEMENT_UNREF(element);
	if (error) {
	    printf("eval error: %s\n", error->message);
	    g_error_free(error);
	    error = NULL;
	}
	if (!el1) continue;
	if (NXP_ELEMENT_GET_TYPE(el1) != NXP_ELEMENT_TYPE_VOID)
	{
	    if ( !(result = nxp_element_repr(el1))) {
		NXP_ELEMENT_UNREF(el1);
		continue;
	    }
	    nxp_user_variable_define_simple("ans", el1, main_context, NULL);
	    printf("ans = %s\n", result->str);
	    g_string_free(result, TRUE);
	}
	NXP_ELEMENT_UNREF(el1);
    } while (line);

    /* --- cleanup and exit --- */
    clear_history();
    nxp_eval_context_destroy(main_context); main_context = NULL;
    nxp_elements_destroy();
    printf("\n");
    return 0;
}



