%{
/* system includes */
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "nxp_kernel.h"
#include "parser.h"
#include "i18n.h"

#define YYERROR_VERBOSE

%}


/*
  This union represents the data that circulates behind the
  non terminal symbols.
*/
%union {
  GHashTable *hash;
  NxpElement *el;
  NxpElement *atom;
  NxpOp f;
  GPtrArray *array;
  char* str;
  NxpIdentifier *id;
  NxpElement *list;
  NxpArg args;
}


/*
  These are the nonterminal symbols, so, here we say what type
  each of the nonterminal symbols has.
 */

%type <atom> INT
%type <atom> REAL
%type <atom> COMPLEX
%type <atom> STRING
%type <atom> function
%type <atom> variable
%type <atom> vector
%type <atom> matrix
%type <id> IDENTIFIER
%type <id> FUNC
%type <id> FUNC1
%type <el> expression
%type <el> expr
%type <el> power
%type <el> product
%type <el> sum
%type <el> atom
%type <el> func
%type <el> line
%type <el> block
%type <array> vector_line
%type <array> matrix_line
%type <args> arg_list

/*
  Here, we define the terminal symbols. They are terminals, so
  they do not need data structures (well, some of them)
*/
%token INT IDENTIFIER FUNC FUNC1
%token REAL COMPLEX 
%token IF THEN ELSE WHILE DO FOR IN
%token STRING
%left  NEWLINE PVIRG
%token LEFT_PAR RIGHT_PAR
%token LEFT_RECT RIGHT_RECT
%token RIGHT_CHAV
%token VIRG
%token RANGE

/* Operators Precedence */
%right OP_ATRIB OP_STORE
%right EQ_GT

%left LT LE GT GE EQ NE
%left OP_PLUS OP_MINUS
%left OP_TIMES OP_OVER OP_PERCENT
%left AND OR NOT
%right UMINUS
%left FACTORIAL OP_INC OP_DEC
%right OP_CIRC SIZE
%left OP_SETA LEFT_CHAV


/* And let the party begin */

%start line

%%

line: block {
    NxpElementList *el = NXP_ELEMENT_LIST($1);
    if (el->list) {
	if (el->list->next) {
	    exp_expression = nxp_element_function_new_simple(":", NXP_ELEMENT(el));
	    NXP_ELEMENT_UNREF((NxpElement *)el);
	}
	else {
	    exp_expression = NXP_ELEMENT_REF(el->list->data);
	    NXP_ELEMENT_UNREF((NxpElement *)el);
	}
    }
    else {
	NXP_ELEMENT_UNREF(el);
	exp_expression = NULL;
    }
}
;

/* An expression or a block. We need that because in a function like "if", we can have just an expression to execute if the condition is true but we can have all a block too. */

/* An exp_expression */
expression:
sum				{ $$ = $1; }
// Comparasions and Logic Operators
| expression LT expression  { $$ = elem_LT($1,$3);  }
| expression LE expression  { $$ = elem_LE($1,$3);  }
| expression GE expression  { $$ = elem_GE($1,$3);  }
| expression GT expression  { $$ = elem_GT($1,$3);  }
| expression NE expression  { $$ = elem_NE($1,$3);  }
| expression EQ expression  { $$ = elem_EQ($1,$3);  }
| expression AND expression { $$ = elem_AND($1,$3); }
| expression OR expression  { $$ = elem_OR($1,$3);  }
| NOT expression           { $$ = elem_NOT($2);    }
| LEFT_RECT block RIGHT_RECT		{
    $$ = nxp_element_function_new_simple(":", $2);
    NXP_ELEMENT_UNREF($2);
}
// The factorial
| expression OP_ATRIB expression {
    $$ = nxp_element_tree_new(NXP_OP_EQ, $1, $3);
    NXP_ELEMENT_UNREF($1);
    NXP_ELEMENT_UNREF($3);
}
| expression OP_ATRIB error {
    $$ = NULL;
    nxp_parser_error(_("Error parsing right side of assignment")); 
    YYABORT;
}
| expression OP_STORE expression {
    NxpElement *tree;

    tree = nxp_element_tree_new(NXP_OP_COLON_EQ, $1, $3);
    NXP_ELEMENT_UNREF($1);
    NXP_ELEMENT_UNREF($3);
    $$ = tree;
}
| expression OP_STORE error {
    $$ = NULL;
    NXP_ELEMENT_UNREF($1);
    nxp_parser_error(_("Error parsing right side of store operation"));
    YYABORT;
}
| SIZE expression {
    $$ = elem_SIZE($2);
}
/*
| expression OP_PLUS error		{
    $$ = NULL;
    nxp_parser_error(_("Error parsing right side of sum")); 
    YYABORT;
}
| expression OP_MINUS error		{
    $$ = NULL;
    nxp_parser_error(_("Error parsing right side of subtraction")); 
    YYABORT;
}
| expression OP_OVER error		{
    $$ = NULL;
    nxp_parser_error(_("Error parsing right side of division")); 
    YYABORT;
}
| expression OP_PERCENT error		{
    $$ = NULL;
    nxp_parser_error(_("Error parsing right side of modulo operator")); 
    YYABORT;
}
| expression OP_CIRC error		{ 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of exponentiation")); 
    YYABORT;
}
| expression LT error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of < comparison"));
    YYABORT;
}
| expression LE error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of <= comparison "));
    YYABORT;
}
| expression GE error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of >= comparison "));
    YYABORT;
}
| expression GT error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of > comparison "));
    YYABORT;
}
| expression NE error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of != comparison "));
    YYABORT;
}
| expression EQ error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of == comparison "));
    YYABORT;
}
| expression AND error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of /\\ operator "));
    YYABORT;
}
| expression OR error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing right side of \\/ operator"));
    YYABORT;
}
| NOT error            { 
    $$ = NULL; 
    nxp_parser_error(_("Error parsing ~ operator argument"));
    YYABORT;
}*/
;

sum:
  product   { $$ = $1; }
| OP_MINUS product { $$ = elem_UMINUS($2); }
| sum OP_PLUS product { $$ = elem_PLUS($1, $3); }
| sum OP_MINUS product { $$ = elem_MINUS($1, $3); }
;

product:
  power	{ $$ = $1; }
| product OP_TIMES power { $$ = elem_TIMES($1, $3); }
| product OP_OVER power { $$ = elem_OVER($1, $3); }
| product OP_PERCENT power { $$ = elem_PERCENT($1, $3); }
| product power { $$ = elem_TIMES($1, $2); }
;

power:
  expr { $$ = $1; }
| power OP_CIRC expr { $$ = elem_CIRC($1, $3); }
;

expr:
  atom { $$ = $1; }
| LEFT_PAR expression RIGHT_PAR { $$ = $2; }
| expr FACTORIAL { 
    NxpElement_t* result;
    result = elem_FACTORIAL($1); 
    $$ = result;
}
| expr OP_SETA expr
{
    NxpElement *result;
    result = nxp_element_container_new($1, $3);
    $$ = result;
}
| expr OP_SETA LEFT_PAR expr VIRG expr RIGHT_PAR
{
    NxpElement* result;
    result = elem_GET_FROM_MATRIX($1,$4,$6); 
    $$ = result;
}
;

block:
NEWLINE {
    $$ = nxp_element_list_new();
}
| expression NEWLINE { $$ = nxp_element_list_new_simple($1); NXP_ELEMENT_UNREF($1); }
| expression PVIRG { $$ = nxp_element_list_new_simple($1); NXP_ELEMENT_UNREF($1); }
| block NEWLINE
| block expression NEWLINE { nxp_element_list_append($1, $2); NXP_ELEMENT_UNREF($2); }
| block expression PVIRG { nxp_element_list_append($1, $2); NXP_ELEMENT_UNREF($2); }
;

/*
  This matrix_line define all that is between '|' in a matrix.  So, a
  matrix is compound by n vectors lines between '|' and ';'. This
  define it recursively.
*/
matrix_line: vector_line PVIRG vector_line {
    GPtrArray* array;
    array = g_ptr_array_new();
    g_ptr_array_add(array, $1);
    g_ptr_array_add(array, $3);
    $$ = array;
}
| matrix_line PVIRG vector_line {
    g_ptr_array_add($1, $3);
    $$ = $1;
};

/*
  This define an atom. An atom (yes, really, it isn't an
  atom) is a real, a funcion, a var, a function times an atom,
  a real times a function or a funcion times a real.
 */
atom:
  INT		{ $$ = $1; }
|  REAL          { $$ = $1; }
| COMPLEX       { $$ = $1; }
| STRING	{ $$ = $1; }
| function	{ $$ = $1; }
| variable 	{ $$ = $1; }
| vector	{ $$ = $1; }
| matrix        { $$ = $1; }
;


arg_list: expression {    
    NxpArg args;
    args.list = nxp_element_list_new_simple($1);
    args.named_arg = g_hash_table_new_full(g_str_hash, g_str_equal,
					   g_free, nxp_element_data_unref);
    NXP_ELEMENT_UNREF($1);
    $$ = args;
}
| arg_list VIRG expression {
    nxp_element_list_append(NXP_ELEMENT($1.list), $3);
    NXP_ELEMENT_UNREF($3);
    $$ = $1;
}
| arg_list VIRG IDENTIFIER EQ_GT expression {
    char *named_arg = nxp_identifier_to_string($3);
    nxp_identifier_unref($3);
    g_hash_table_insert($1.named_arg, named_arg, $5);
    $$ = $1;
}
;

/* This defines built-in's or user defined functions. */
function: func {$$ = $1;}
| IF expression THEN LEFT_RECT block RIGHT_RECT {
    NxpElement *expr, *args;
    expr = nxp_element_function_new_simple(":", $5);
    NXP_ELEMENT_UNREF($5);
    args = nxp_element_list_new();
    nxp_element_list_append(args, $2);
    NXP_ELEMENT_UNREF($2);
    nxp_element_list_append(args, expr);
    NXP_ELEMENT_UNREF(expr);
    $$ = nxp_element_function_new_simple("if", args);
    NXP_ELEMENT_UNREF(args);
}
| IF expression THEN LEFT_RECT block RIGHT_RECT ELSE LEFT_RECT block RIGHT_RECT {
    NxpElement *expr1, *expr2, *args;
    expr1 = nxp_element_function_new_simple(":", $5);
    NXP_ELEMENT_UNREF($5);
    expr2 = nxp_element_function_new_simple(":", $9);
    NXP_ELEMENT_UNREF($9);
    args = nxp_element_list_new_simple($2);
    NXP_ELEMENT_UNREF($2);
    nxp_element_list_append(args, expr1);
    NXP_ELEMENT_UNREF(expr1);
    nxp_element_list_append(args, expr2);
    NXP_ELEMENT_UNREF(expr2);
    $$ = nxp_element_function_new_simple("if", args);
    NXP_ELEMENT_UNREF(args);
}
| IF LEFT_RECT expression VIRG expression RIGHT_RECT {
    NxpElement *args;
    args = nxp_element_list_new_simple($3);
    nxp_element_list_append(args, $5);
    NXP_ELEMENT_UNREF($3);
    NXP_ELEMENT_UNREF($5);
    $$ = nxp_element_function_new_simple("if", args);
    NXP_ELEMENT_UNREF(args);
}
| IF LEFT_RECT expression VIRG expression VIRG expression RIGHT_RECT {
    NxpElement *args;
    args = nxp_element_list_new_simple($3);
    nxp_element_list_append(args, $5);
    nxp_element_list_append(args, $7);
    NXP_ELEMENT_UNREF($3);
    NXP_ELEMENT_UNREF($5);
    NXP_ELEMENT_UNREF($7);
    $$ = nxp_element_function_new_simple("if", args);
    NXP_ELEMENT_UNREF(args);
}| FOR variable IN expression DO LEFT_RECT block RIGHT_RECT {
    NxpElement *args, *expr;
    args = nxp_element_list_new_simple($2);
    NXP_ELEMENT_UNREF($2);
    nxp_element_list_append(args, $4);
    NXP_ELEMENT_UNREF($4);
    expr = nxp_element_function_new_simple(":", $7);
    NXP_ELEMENT_UNREF($7);
    nxp_element_list_append(args, expr);
    NXP_ELEMENT_UNREF(expr);
    $$ = nxp_element_function_new_simple("for", args);
    NXP_ELEMENT_UNREF(args);
}| FOR LEFT_RECT variable VIRG expression VIRG expression RIGHT_RECT {
    NxpElement *args;
    args = nxp_element_list_new();
    nxp_element_list_append(args, $3);
    NXP_ELEMENT_UNREF($3);
    nxp_element_list_append(args, $5);
    NXP_ELEMENT_UNREF($5);
    nxp_element_list_append(args, $7);
    NXP_ELEMENT_UNREF($7);
    $$ = nxp_element_function_new_simple("for", args);
    NXP_ELEMENT_UNREF(args);
}| WHILE expression DO LEFT_RECT block RIGHT_RECT {
    NxpElement *args;
    args = nxp_element_list_new_simple($2);
    NXP_ELEMENT_UNREF($2);
    nxp_element_list_append_list(args, $5);
    NXP_ELEMENT_UNREF($5);
    $$ = nxp_element_function_new_simple("while", args);
    NXP_ELEMENT_UNREF(args);
}| WHILE LEFT_RECT expression VIRG arg_list RIGHT_RECT {
    NxpElement *args;
    args = nxp_element_list_new_simple($3);
    nxp_element_list_append_list(args, $5.list);
    NXP_ELEMENT_UNREF($3);
    NXP_ELEMENT_UNREF($5.list);
    $$ = nxp_element_function_new_simple("while", args);
    NXP_ELEMENT_UNREF(args);
};

func:
FUNC arg_list RIGHT_RECT {
    $$ = nxp_element_function_new($1, $2.list, $2.named_arg);
    NXP_ELEMENT_UNREF($2.list);
    nxp_identifier_unref($1);
}
| FUNC1 arg_list RIGHT_PAR {
    $$ = nxp_element_function_new($1, $2.list, $2.named_arg);
    NXP_ELEMENT_UNREF($2.list);
    nxp_identifier_unref($1);
}
/* Or a empty one */
| FUNC RIGHT_RECT {
    NxpElement *list = nxp_element_list_new();
    $$ = nxp_element_function_new($1, list, NULL);
    NXP_ELEMENT_UNREF(list);
    nxp_identifier_unref($1);
}
| FUNC1 RIGHT_PAR {
    NxpElement *list = nxp_element_list_new();
    $$ = nxp_element_function_new($1, list, NULL);
    NXP_ELEMENT_UNREF(list);
    nxp_identifier_unref($1);
}
;

/* Here, we recognize a vector. That is, { line } */
vector: LEFT_CHAV vector_line RIGHT_CHAV {
    int pos, size;
    NxpElementVector *vec;
    NxpElement *el;

    size = $2->len;
    vec = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (pos = 0; pos < size; pos++) {
	el = g_ptr_array_index($2, pos);
	_element_vector_index(vec, pos) = el;
    }
    g_ptr_array_free($2, TRUE);

    $$ = NXP_ELEMENT(vec);
}
| LEFT_CHAV  expression RANGE expression RIGHT_CHAV {
    $$ = elem_DOT3($2,$4); 
}
| LEFT_CHAV RIGHT_CHAV {
    $$ = nxp_element_vector_new(0, NULL);
}
;

/* a line from a vector is an expression or a comma separated list of
   expressions */
vector_line : expression {
    GPtrArray* array;
    array = g_ptr_array_new();
    g_ptr_array_add(array, $1);
    $$ = array;
}
| vector_line VIRG expression {
    g_ptr_array_add($1, $3);
    $$ = $1;
};

matrix: LEFT_CHAV matrix_line RIGHT_CHAV {
    guint i, j, height, width, pmax;
    NxpElementMatrix *m;
    GPtrArray *mline;
    
    mline = $2;
    height = mline->len;		/* height of matrix */
    /* get maximum of widths of each row in the matrix */
    width = 0;

    for (i = 0; i < height; i++) {
	pmax = ((GPtrArray*) (g_ptr_array_index(mline, i)))->len;
	if (pmax > width)
	    width = pmax;
    }

    m = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(width, height, NULL));
    for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
	    if (j < ((GPtrArray*)g_ptr_array_index(mline, i))->len)
		_element_matrix_index(m, i, j) =  
		    g_ptr_array_index(
			((GPtrArray*)g_ptr_array_index(mline, i)), j);
	    else
		_element_matrix_index(m, i, j) =  nxp_element_int_new(0);
    
    for (i = 0; i < height; i++)
	g_ptr_array_free((GPtrArray*)g_ptr_array_index(mline, i), TRUE);
    g_ptr_array_free(mline, TRUE);

    $$ = NXP_ELEMENT(m);
};

variable:
IDENTIFIER %prec OP_ATRIB {
    $$ = nxp_element_variable_new($1);
    nxp_identifier_unref($1);
}
| OP_INC variable {
    $$ = nxp_element_tree_new(NXP_OP_PLUS2_LEFT, NULL, $2);
    NXP_ELEMENT_UNREF($2);
}
| OP_DEC variable {
    $$ = nxp_element_tree_new(NXP_OP_MINUS2_LEFT, NULL, $2);
    NXP_ELEMENT_UNREF($2);
}
;

/* End of productions */
%%

/* END OF FILE */
